from __future__ import annotations

from functools import cached_property
from typing import Any

from plain.assets.views import AssetView
from plain.http import (
    Http404,
    Response,
    ResponseRedirect,
)
from plain.runtime import settings
from plain.views import TemplateView, View

from .exceptions import PageNotFoundError, RedirectPageError
from .pages import Page
from .registry import pages_registry


class PageViewMixin:
    @cached_property
    def page(self) -> Page:
        url_name = self.request.resolver_match.url_name

        try:
            return pages_registry.get_page_from_name(url_name)
        except PageNotFoundError:
            raise Http404()


class PageView(PageViewMixin, TemplateView):
    template_name = "page.html"

    def get(self) -> Response:
        """Check Accept header and serve markdown if requested."""
        if self.page.is_markdown() and settings.PAGES_SERVE_MARKDOWN:
            preferred = self.request.get_preferred_type(
                "text/markdown", "text/plain", "text/html"
            )
            if preferred in ("text/markdown", "text/plain"):
                markdown_content = self.page._frontmatter.content
                response = Response(
                    markdown_content, content_type="text/plain; charset=utf-8"
                )
                response.headers["Vary"] = "Accept"
                return response

        return super().get()

    def get_template_names(self) -> list[str]:
        """
        Allow for more specific user templates like
        markdown.html or html.html
        """
        if template_name := self.page.get_template_name():
            return [template_name]

        return super().get_template_names()

    def get_template_context(self) -> dict[str, Any]:
        context = super().get_template_context()
        context["page"] = self.page
        self.page.set_template_context(context)  # Pass the standard context through
        return context


class PageRedirectView(PageViewMixin, View):
    def get(self) -> ResponseRedirect:
        url = self.page.vars.get("url")

        if not url:
            raise RedirectPageError("Redirect page is missing a url")

        status_code = self.page.vars.get("status_code", 302)
        return ResponseRedirect(url, status_code=status_code)


class PageAssetView(PageViewMixin, AssetView):
    def get_url_path(self) -> str | None:
        return self.page.get_url_path()

    def get_asset_path(self, path: str) -> str:
        return self.page.absolute_path

    def get_debug_asset_path(self, path: str) -> str:
        return self.page.absolute_path


class PageMarkdownView(PageViewMixin, View):
    def get(self) -> Response:
        """Serve the markdown content without frontmatter."""
        markdown_content = self.page._frontmatter.content
        response = Response(markdown_content, content_type="text/plain; charset=utf-8")
        response.headers["Vary"] = (
            "Accept-Encoding"  # Set Vary header for proper caching
        )
        return response



# Enable serving raw markdown content (without frontmatter)
# When enabled, markdown pages can be served as raw markdown via:
# - Accept header negotiation (Accept: text/markdown or text/plain)
# - Separate .md URLs (e.g., /docs/guide.md alongside /docs/guide/)
PAGES_SERVE_MARKDOWN = False



from __future__ import annotations

import os
from html.parser import HTMLParser
from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse, urlunparse

import mistune
from pygments import highlight
from pygments.formatters import html
from pygments.lexers import get_lexer_by_name

from plain.urls import reverse
from plain.utils.text import slugify

if TYPE_CHECKING:
    from .registry import PagesRegistry


class PagesRenderer(mistune.HTMLRenderer):
    def __init__(
        self, current_page_path: str, pages_registry: PagesRegistry, **kwargs: Any
    ):
        super().__init__(**kwargs)
        self.current_page_path = current_page_path
        self.pages_registry = pages_registry

    def link(self, text: str, url: str, title: str | None = None) -> str:
        """Convert relative markdown links to proper page URLs."""
        # Check if it's a relative link (starts with ./ or ../, or is just a filename)
        is_relative = url.startswith(("./", "../")) or (
            not url.startswith(("http://", "https://", "/", "#")) and ":" not in url
        )

        if is_relative:
            # Parse URL to extract components
            parsed_url = urlparse(url)

            # Resolve relative to current page's directory using just the path component
            current_dir = os.path.dirname(self.current_page_path)
            resolved_path = os.path.normpath(os.path.join(current_dir, parsed_url.path))
            page = self.pages_registry.get_page_from_path(resolved_path)

            # Get the primary URL name for link conversion
            url_name = page.get_url_name()
            if url_name:
                base_url = reverse(f"pages:{url_name}")
                # Reconstruct URL with preserved query params and fragment
                url = str(
                    urlunparse(
                        (
                            parsed_url.scheme,  # scheme (empty for relative)
                            parsed_url.netloc,  # netloc (empty for relative)
                            base_url,  # path (our converted URL)
                            parsed_url.params,  # params
                            parsed_url.query,  # query
                            parsed_url.fragment,  # fragment
                        )
                    )
                )

        return super().link(text, url, title)

    def heading(self, text: str, level: int, **attrs: Any) -> str:
        """Automatically add an ID to headings if one is not provided."""

        if "id" not in attrs:
            inner_text = get_inner_text(text)
            inner_text = inner_text.replace(
                ".", "-"
            )  # Replace dots with hyphens (slugify won't)
            attrs["id"] = slugify(inner_text)

        return super().heading(text, level, **attrs)

    def block_code(self, code: str, info: str | None = None) -> str:
        """Highlight code blocks using Pygments."""

        if info:
            lexer = get_lexer_by_name(info, stripall=True)
            formatter = html.HtmlFormatter(wrapcode=True)
            return highlight(code, lexer, formatter)

        return "<pre><code>" + mistune.escape(code) + "</code></pre>"


def render_markdown(content: str, current_page_path: str) -> str:
    from .registry import pages_registry

    renderer = PagesRenderer(
        current_page_path=current_page_path, pages_registry=pages_registry, escape=False
    )
    markdown = mistune.create_markdown(
        renderer=renderer, plugins=["strikethrough", "table"]
    )
    return markdown(content)  # type: ignore[return-value]


class InnerTextParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.text_content: list[str] = []

    def handle_data(self, data: str) -> None:
        # Collect all text data
        self.text_content.append(data.strip())


def get_inner_text(html_content: str) -> str:
    parser = InnerTextParser()
    parser.feed(html_content)
    return " ".join([text for text in parser.text_content if text])



from __future__ import annotations

import os
from functools import cached_property
from typing import Any

import frontmatter

from plain.runtime import settings
from plain.templates import Template
from plain.urls import URLPattern, path, reverse

from .markdown import render_markdown


class PageRenderError(Exception):
    pass


class Page:
    def __init__(self, relative_path: str, absolute_path: str):
        self.relative_path = relative_path
        self.absolute_path = absolute_path
        self._template_context: dict[str, Any] = {}

    def set_template_context(self, context: dict[str, Any]) -> None:
        self._template_context = context

    @cached_property
    def _frontmatter(self) -> Any:
        with open(self.absolute_path) as f:
            return frontmatter.load(f)

    @cached_property
    def vars(self) -> dict[str, Any]:
        return self._frontmatter.metadata

    @cached_property
    def title(self) -> str:
        default_title = os.path.splitext(os.path.basename(self.relative_path))[0]
        return self.vars.get("title", default_title)

    @cached_property
    def content(self) -> str:
        # Strip the frontmatter
        content = self._frontmatter.content

        if not self.vars.get("render_plain", False):
            template = Template(os.path.join("pages", self.relative_path))

            try:
                content = template.render(self._template_context)
            except Exception as e:
                # Throw our own error so we don't get shadowed by the Jinja error
                raise PageRenderError(f"Error rendering page {self.relative_path}: {e}")

            # Strip the frontmatter again, since it was in the template file itself
            _, content = frontmatter.parse(content)

        if self.is_markdown():
            content = render_markdown(content, current_page_path=self.relative_path)

        return content

    def is_markdown(self) -> bool:
        extension = os.path.splitext(self.absolute_path)[1]
        return extension == ".md"

    def is_template(self) -> bool:
        return ".template." in os.path.basename(self.absolute_path)

    def is_asset(self) -> bool:
        extension = os.path.splitext(self.absolute_path)[1]
        # Anything that we don't specifically recognize for pages
        # gets treated as an asset
        return extension.lower() not in (
            ".html",
            ".md",
            ".redirect",
        )

    def is_redirect(self) -> bool:
        extension = os.path.splitext(self.absolute_path)[1]
        return extension == ".redirect"

    def get_template_name(self) -> str:
        if template_name := self.vars.get("template_name"):
            return template_name

        return ""

    def get_url_path(self) -> str | None:
        """Generate the primary URL path for this page."""
        if self.is_template():
            return None

        if self.is_asset():
            return self.relative_path

        url_path = os.path.splitext(self.relative_path)[0]

        # If it's an index.html or something, the url is the parent dir
        if os.path.basename(url_path) == "index":
            url_path = os.path.dirname(url_path)

        # The root url should stay an empty string
        if not url_path:
            return ""

        # Everything else should get a trailing slash
        return url_path + "/"

    def get_url_name(self) -> str | None:
        """Generate the URL name from the URL path."""
        url_path = self.get_url_path()
        if url_path is None:
            return None

        if not url_path:
            return "index"

        return url_path.rstrip("/")

    def get_view_class(self) -> type:
        """Get the appropriate view class for this page."""
        from .views import PageAssetView, PageRedirectView, PageView

        if self.is_redirect():
            return PageRedirectView

        if self.is_asset():
            return PageAssetView

        return PageView

    def get_markdown_url(self) -> str | None:
        """Get the markdown URL for this page if it exists."""
        if not self.is_markdown():
            return None

        if not settings.PAGES_SERVE_MARKDOWN:
            return None

        url_name = self.get_url_name()
        if not url_name:
            return None

        return reverse(f"pages:{url_name}-md")

    def get_urls(self) -> list[URLPattern]:
        """Get all URL path objects for this page."""
        urls = []

        # Generate primary URL
        url_path = self.get_url_path()
        url_name = self.get_url_name()
        view_class = self.get_view_class()

        if url_path is not None and url_name is not None:
            urls.append(
                path(
                    url_path,
                    view_class,
                    name=url_name,
                )
            )

            # For markdown files, optionally add .md URL
            if self.is_markdown() and settings.PAGES_SERVE_MARKDOWN:
                from .views import PageMarkdownView

                urls.append(
                    path(
                        self.relative_path,
                        PageMarkdownView,
                        name=f"{url_name}-md",
                    )
                )

        return urls
